#ifndef _BF_IA32OPS_H_
#define _BF_IA32OPS_H_

/*
 * IA32 ISA utilities for brainf*.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * (C) 2008 by Thalion Graphic Solutions eK.
 */

/* register numbers */
#define EAX 0x0
#define ECX 0x1
#define EDX 0x2
#define EBX 0x3

#define ESP 0x4
#define EBP 0x5
#define ESI 0x6
#define EDI 0x7


/* mod r/m fields */
#define MOD_REG_NODIS 0x0
#define MOD_REG_DISPB 0x1
#define MOD_REG_DISPW 0x2
#define MOD_REG_ID    0x3

typedef union {
	unsigned char byte;
	struct {
		unsigned reg1:3;
		unsigned reg2:3;
		unsigned mod: 2;
	} bits;
} modrm_t;

typedef union {
	unsigned char byte;
	struct {
		unsigned base:  3;
		unsigned index: 3;
		unsigned scale: 2;
	} bits;
} sib_t;

#define ADD_I(cp,imm,reg) {\
	modrm_t mod;\
	mod.bits.mod = MOD_REG_ID;\
	mod.bits.reg2 = 0;\
	mod.bits.reg1 = (reg);\
	*(char *)cp++ = 0x81;\
	*(char *)cp++ = mod.byte;\
	*(unsigned int *)cp = (imm);\
	cp += 4;\
}

#define SUB_I(cp,imm,reg) {\
	modrm_t mod;\
	mod.bits.mod = MOD_REG_ID;\
	mod.bits.reg2 = 5;\
	mod.bits.reg1 = (reg);\
	*(char *)cp++ = 0x81;\
	*(char *)cp++ = mod.byte;\
	*(unsigned int *)cp = (imm);\
	cp += 4;\
}

#define MOV_RR(cp,src,dest) {\
	modrm_t mod;\
	mod.bits.mod = MOD_REG_ID;\
	mod.bits.reg2 = (src);\
	mod.bits.reg1 = (dest);\
	*(char *)cp++ = 0x89;\
	*(char *)cp++ = mod.byte;\
}

#define MOV_RI(cp,reg,imm) {\
	*(char *)cp++ = 0xb8 + (reg);\
}

#define MOV_PAC(cp,imm) {\
	*(char *)cp++ = 0xa1;\
	*(unsigned int *)cp = (imm);\
	cp += 4;\
}

#define MOV_ACP(cp,imm) {\
	*(char *)cp++ = 0xa3;\
	*(unsigned int *)cp = (imm);\
	cp += 4;\
}

#define MOV_MR(cp,src,dest,disp) {\
	modrm_t mod;\
	mod.bits.reg2 = (dest);\
	mod.bits.reg1 = (src);\
	if ( (disp) > 0xff) {\
		mod.bits.mod = MOD_REG_DISPW;\
		*(char *)cp++ = 0x8b;\
		*(char *)cp++ = mod.byte;\
		*(unsigned int *)cp = (disp);\
		cp += 4;\
	} \
	else if ( (disp) ) {\
		mod.bits.mod = MOD_REG_DISPB;\
		*(char *)cp++ = 0x8b;\
		*(char *)cp++ = mod.byte;\
		*(char *)cp++ = (disp);\
	}\
	else {\
		mod.bits.mod = MOD_REG_NODIS;\
		*(char *)cp++ = 0x8b;\
		*(char *)cp++ = mod.byte;\
	}\
}

#define MOV_RM(cp,src,dest,disp) {\
	modrm_t mod;\
	mod.bits.reg2 = (src);\
	mod.bits.reg1 = (dest);\
	if ( (disp) > 0xff ) {\
		mod.bits.mod = MOD_REG_DISPW;\
		*(char *)cp++ = 0x89;\
		*(char *)cp++ = mod.byte;\
		*(unsigned int *)cp = (disp);\
		cp += 4;\
	} \
	else if ( (disp) ) {\
		mod.bits.mod = MOD_REG_DISPB;\
		*(char *)cp++ = 0x89;\
		*(char *)cp++ = mod.byte;\
		*(char *)cp++ = (disp);\
	}\
	else {\
		mod.bits.mod = MOD_REG_NODIS;\
		*(char *)cp++ = 0x89;\
		*(char *)cp++ = mod.byte;\
		if ( ESP == (dest) ) {\
			sib_t sib;\
			sib.bits.base = ESP;\
			sib.bits.index = ESP;\
			sib.bits.scale = 0;\
			*(char *)cp++ = sib.byte;\
		}\
	}\
}

#define PUSH(cp,reg) {\
	*(char *)cp++ = 0x50 + (reg);\
}

#define POP(cp,reg) {\
	*(char *)cp++ = 0x58 + (reg);\
}

#define CALL(cp,addr) {\
	unsigned int *disp;\
	*(char *)cp++ = 0xe8;\
	disp = (unsigned int *)cp; cp += 4; \
	*disp = ((unsigned int)&(addr))-(unsigned int)cp;\
}

#define CALL_R(cp,reg) {\
	modrm_t mod;\
	mod.bits.mod = MOD_REG_ID;\
	mod.bits.reg1 = (reg);\
	mod.bits.reg2 = 2;\
	*(char *)cp++ = 0xFF;\
	*(char *)cp++ = mod.byte;\
}

#define CALL_M(cp,mem) {\
	modrm_t mod;\
	mod.bits.mod = MOD_REG_NODIS;\
	mod.bits.reg1 = 5;\
	mod.bits.reg2 = 2;\
	*(char *)cp++ = 0xFF;\
	*(char *)cp++ = mod.byte;\
}

#define CMP_I(cp,imm) {\
	*(char *)cp++ = 0x3d;\
	*(unsigned int *)cp = (imm);\
	cp += 4;\
}

#define LEAVE(cp) {\
	*(char *)cp++ = 0xc9;\
}
	
#define RET(cp) {\
	*(char *)cp++ = 0xC3;\
}

#define JECXZ(cp,label) {\
	*(char *)cp++ = 0xe3;\
	*(char *)cp++ = (label);\
}

#define JZ(cp,label) {\
	*(char *)cp++ = 0x0f;\
	*(char *)cp++ = 0x84;\
	*(unsigned int*)cp = (label);\
	cp += 4;\
}

#define JNZ(cp,label) {\
	*(char *)cp++ = 0x0f;\
	*(char *)cp++ = 0x85;\
	*(unsigned int*)cp = (label);\
	cp += 4;\
}

#endif
