#ifndef _BF_MIPSOPS_H_
#define _BF_MIPSOPS_H_

/*
 * Mips ISA utilities for brainf*.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * (C) 2003 by Thalion Graphic Solutions eK.
 */

typedef struct {
	unsigned op_:6;
	unsigned rs_:5;
	unsigned rt_:5;
	unsigned immediate_:16;
} mips_itype_t;

typedef struct {
	unsigned op_:6;
	unsigned target_:26;
} mips_jtype_t;

typedef struct {
	unsigned op_:6;
	unsigned rs_:5;
	unsigned rt_:5;
	unsigned rd_:5;
	unsigned shamt_:5;
	unsigned funct_:6;
} mips_rtype_t;

typedef union {

	mips_itype_t itype_;
	mips_jtype_t jtype_;
	mips_rtype_t rtype_;

} mips_ins_t;

#define ADDI(cp,rs,rt,immediate) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x8;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->immediate_ = (immediate);\
	(cp)++;\
}

#define ADDIU(cp,rs,rt,immediate) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x9;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->immediate_ = (immediate);\
	(cp)++;\
}

#define ADDU(cp,rs,rt,rd) {\
	mips_rtype_t *ins = (mips_rtype_t *)cp;\
	ins->op_ = 0x0;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->rt_ = (rd);\
	ins->shamt_ = 0;\
	ins->funct_ = 0x21;\
	(cp)++;\
}

#define BEQ(cp,rs,rt,offset) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x4;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->immediate_ = (offset);\
	(cp)++;\
}

#define BNE(cp,rs,rt,offset) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x5;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->immediate_ = (offset);\
	(cp)++;\
}

#define LD(cp,base,rt,offset) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x37;\
	ins->rs_ = (base);\
	ins->rt_ = (rt);\
	ins->immediate_ = (offset);\
	(cp)++;\
}

#define LW(cp,base,rt,offset) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x23;\
	ins->rs_ = (base);\
	ins->rt_ = (rt);\
	ins->immediate_ = (offset);\
	(cp)++;\
}

#define JALR(cp,rs,rd) {\
	mips_rtype_t *ins = (mips_rtype_t *)cp;\
	ins->op_ = 0x0;\
	ins->rs_ = (rs);\
	ins->rt_ = 0;\
	ins->rd_ = (rd);\
	ins->shamt_ = 0;\
	ins->funct_ = 0x9;\
	(cp)++;\
}

#define JR(cp,rs) {\
	mips_rtype_t *ins = (mips_rtype_t *)cp;\
	ins->op_ = 0;\
	ins->rs_ = (rs);\
	ins->rt_ = 0;\
	ins->rd_ = 0;\
	ins->shamt_ = 0;\
	ins->funct_ = 0x8;\
	(cp)++;\
}

#define LI(cp,rs,rt,immediate) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x8;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->immediate_ = (immediate);\
	(cp)++;\
}


#define LUI(cp,rt,immediate) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0xf;\
	ins->rs_ = 0;\
	ins->rt_ = (rt);\
	ins->immediate_ = (immediate);\
	(cp)++;\
}

#define NOP(cp) {\
	mips_jtype_t *ins = (mips_jtype_t *)cp;\
	ins->op_ = 0;\
	ins->target_ = 0;\
	(cp)++;\
}

#define OR(cp,rs,rt,rd) {\
	mips_rtype_t *ins = (mips_rtype_t *)cp;\
	ins->op_ = 0;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->rd_ = (rd);\
	ins->shamt_ = 0;\
	ins->funct_ = 0x25;\
	(cp)++;\
}

#define ORI(cp,rs,rt,immediate) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0xd;\
	ins->rs_ = (rs);\
	ins->rt_ = (rt);\
	ins->immediate_ = (immediate);\
	(cp)++;\
}

#define SD(cp,base,rt,offset) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x3f;\
	ins->rs_ = (base);\
	ins->rt_ = (rt);\
	ins->immediate_ = (offset);\
	(cp)++;\
}

#define SW(cp,base,rt,offset) {\
	mips_itype_t *ins = (mips_itype_t *)cp;\
	ins->op_ = 0x2b;\
	ins->rs_ = (base);\
	ins->rt_ = (rt);\
	ins->immediate_ = (offset);\
	(cp)++;\
}

#define T9 25
#define SP 29
#define GP 28
#define FP 30
#define RA 31

#define hi(addr) (((uint32_t)addr)>>16)
#define lo(addr) (((uint32_t)addr)&0xffff)

#endif
