/*
 * A brainf* JIT for the SGI Irix/MIPS-platform.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * (C) 2003 by Thalion Graphic Solutions eK.
 */
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ucontext.h>

#include <bf.h>

#define PAGESIZ sysconf(_SC_PAGESIZE)
#define DEFAULT_CS_NPAGES 4
#define DEFAULT_DS_NPAGES 5
#define MIN_CS_NPAGES 1
#define MIN_DS_NPAGES 1

typedef int (*func_t)(void *);

static int *ds;
static size_t ds_size;
static size_t cs_size;
static int i;
static char *bf_name;
static char *prog_name;
static FILE *in, *out;

cp_t cs, cp;

static void get(uint32_t *p)
{
	*p = feof(in) ? 0 : fgetc(in);
}

static void put(uint32_t *p)
{
	fputc(*p,out);
}

static void compile_block(FILE *file, int label, int depth)
{
	int c;
	int n = 0;
	while( (c=fgetc(file)) != EOF ) {
		if ( i > cs_size-8 ) {
			fprintf(stderr,"%s: program size exceeds code segment size (%d).\n",bf_name,cs_size);
			fprintf(stderr,"%s: try re-running the program using -c<larger-size>.\n",bf_name);
			exit(-1); 
		}
		switch( c ) {
			case '>':
				ungetc(c,file);
				while ( (c=fgetc(file)) == '>' ) {
					n++;
				}
				ungetc(c,file);
				INCP(n);
				n=0;
				break;
			case '<':
				ungetc(c,file);
				while ( (c=fgetc(file)) == '<' ) {
					n++;
				}
				ungetc(c,file);
				DECP(n);
				n=0;
				break;
			case '+':
				ungetc(c,file);
				while ( (c=fgetc(file)) == '+' ) {
					n++;
				}
				ungetc(c,file);
				INCV(n);
				n=0;
				break;
			case '-':
				ungetc(c,file);
				while ( (c=fgetc(file)) == '-' ) {
					n++;
				}
				ungetc(c,file);
				DECV(n);
				n=0;
				break;
			case '.':
				CALL_V(put);
				break;
			case ',':
				CALL_V(get);
				break;
			case '[':
				BEGIN();
				break;
			case ']':
				if ( depth ) {
					END();
					return;
				}
				else {
					fprintf(stderr,"Unbalanced ']' found\n");
					exit(-1);
				}
				break;
			default:
				break;
				
		}
	}
}

static void compile(FILE *file)
{
	memset(cs,0,cs_size);
	cp = cs;
	i=0;
	COMPILE(file);
}

static void dump()
{
	cp_t *p = cs;
	while ( p<cp ) {
		printf("0x%X\n",*p++);
	}
}


static void handle_segv(int sig,siginfo_t *info,void *p)
{
	ucontext_t *uc = (ucontext_t *)p;
	if ( REG(uc,P) < (greg_t) ds ) {
		fprintf(stderr,"%s: program '%s' data pointer exceeded legal memory area\n",bf_name,prog_name);
		exit(-1);
	}
	if ( REG(uc,P) > (greg_t) (((char *)ds)+ds_size) ) {
		fprintf(stderr,"%s: program '%s' data pointer exceeded legal memory area\n",bf_name,prog_name);
		exit(-1);
	}
	fprintf(stderr,"%s: unexpected sigsegv caught.\n",bf_name);
	exit(-2);
}


static void init(size_t ds_size,size_t cs_size)
{
	struct sigaction action;
	char *p = NULL;
	int i=1, fd = open("/dev/zero", O_RDWR);
	if ( fd < 0 || (p = (char *)mmap(0,cs_size,PROT_READ|PROT_WRITE|PROT_EXEC,MAP_PRIVATE,fd,0)) == (char *)-1 ) {
		fprintf(stderr,"%s: could not allocate code segment.\n",bf_name);
		exit(-1);
	}
	cs = (cp_t *)p;
	ds_size += 2 * PAGESIZ;
	if ( (p = (char *)mmap(0,ds_size+2*PAGESIZ,PROT_READ|PROT_WRITE,MAP_PRIVATE,fd,0)) == (char *)-1 ) {
		fprintf(stderr,"%s: could not allocate data segment.\n",bf_name);
		exit(-1);
	}
	if ( mprotect(p,PAGESIZ,PROT_NONE) ||
	     mprotect(p+ds_size+PAGESIZ,PAGESIZ,PROT_NONE) ) {
		fprintf(stderr,"%s: could not initialize data segment.\n",bf_name);
		exit(-1);
	}
	ds = (int *)(p+PAGESIZ);
	memset(ds,0,ds_size);

	/* Install a sigsegv handler */
	action.sa_flags = SA_SIGINFO;
	action.sa_handler = NULL;
	sigemptyset(&action.sa_mask);
	action.sa_sigaction = handle_segv;
	sigaction(SIGSEGV,&action,NULL);
}

static int run()
{
	func_t f = (func_t) cs;
	if ( mprotect(cs,cs_size,PROT_EXEC) ) {
		fprintf(stderr,"%s: couldn't initialize code segment.\n",bf_name);
		exit(-1);
	}
	return f(ds);
}

static void usage(char *name)
{
	printf("Usage: %s [-h] | [-c<npages>] [-d<npages>] [-i<infile>] <program>\n",name);
}

static void help(char *name)
{
	printf("%s - the Thalion Graphic Solutions eK. brainf* JIT.\n\n",name);
	usage(name);
	printf("-c<npages>   uses <npages>*%d 32bit words as memory area for compiled code\n",PAGESIZ/sizeof(uint32_t));
	printf("             (default is %d)\n",DEFAULT_CS_NPAGES);
	printf("-d<npages>   uses <npages>*%d 32bit words as brainf* memory cell array\n",PAGESIZ/sizeof(int));
	printf("             (default is %d)\n",DEFAULT_DS_NPAGES);
	printf("-i<infile>   uses <infile> instead of stdin for brainf* read operation\n");
	printf("-v           print return value of program\n");
	printf("-h           prints this help\n");
}

int main(int argc,char *argv[])
{
	int c;
	int npages;
	int verbose = 0;
	extern char *optarg;
	extern int optind;
	cs_size=PAGESIZ*DEFAULT_CS_NPAGES;
	ds_size=PAGESIZ*DEFAULT_DS_NPAGES;
	in = stdin;
	out = stdout;
	while ( (c = getopt(argc, argv, "c:d:i:vh")) != -1) {
		switch ( c ) {
			case 'c':
				npages = atoi(optarg);
				if ( npages>MIN_CS_NPAGES ) {
					cs_size=npages*PAGESIZ;
				}
                                break;
			case 'd':
				npages = atoi(optarg);
				if ( npages>MIN_DS_NPAGES ) {
					ds_size=npages*PAGESIZ;
				}
                                break;
			case 'i':
				if ( (in=fopen(optarg,"rb")) == NULL ) { 
					fprintf(stderr,"%s: could not open file '%s'.\n",argv[0],optarg);
					return -1;
				}
                                break;

			case 'v':
				verbose = 1;
				break;
			case 'h':
				help(argv[0]);
				return 0;
			default:
                                printf("%s: invalid option '%c'.\n",argv[0],c);
                                return -1;
                }
        }
	bf_name = argv[0];
	init(ds_size,cs_size);
	if ( argc-optind == 1 ) {
		FILE *file;
		prog_name = argv[optind];
		if ( file = fopen(prog_name,"rb") ) {
			int rv = 0;
			compile(file);
			rv = run();
			if ( verbose )
				printf("return value: %d\n",rv);
		}
		else {
			fprintf(stderr,"%s: could not open file '%s'.\n",argv[0],prog_name);
			return -1;
		}
	}
	else {
		usage(argv[0]);
	}
	return 0;
}


